@@ -143,6 +143,7 @@ struct knot_packet {
 	knot_packet_prealloc_type_t prealloc_type;
 	
 	size_t tsig_size;	/*!< Space to reserve for the TSIG RR. */
+	knot_rrset_t *tsig_rr;  /*!< TSIG RR stored in the packet. */
 };
 
 typedef struct knot_packet knot_packet_t;
@@ -378,6 +379,10 @@ int knot_packet_arcount(const knot_packet_t *packet);
 
 void knot_packet_set_tsig_size(knot_packet_t *packet, size_t tsig_size);
 
+const knot_rrset_t *knot_packet_tsig(const knot_packet_t *packet);
+
+void knot_packet_set_tsig(knot_packet_t *packet, const knot_rrset_t *tsig_rr);
+
 /*!
  * \brief Returns number of RRSets in Answer section of the packet.
  *
@@ -439,7 +444,7 @@ const knot_rrset_t *knot_packet_authority_rrset(
  *         or NULL if there is no such RRSet.
  */
 const knot_rrset_t *knot_packet_additional_rrset(
-	knot_packet_t *packet, short pos);
+    knot_packet_t *packet, short pos);
 
 /*!
  * \brief Checks if the packet already contains the given RRSet.
